""" Load various data formats into xibabel images
"""

from functools import partial
from pathlib import Path
import json
import logging
import importlib
import os.path as op

import numpy as np
import psutil

import nibabel as nib
from nibabel.spatialimages import HeaderDataError
from nibabel.filename_parser import parse_filename
from nibabel.fileholders import FileHolder
from nibabel.affines import from_matvec, apply_affine

import fsspec
import xarray as xr
import dask.array as da

from .xutils import merge


logger = logging.getLogger(__name__)


def max_available_div(div=10):
    """ Set default chunk as fraction of maximum aailable memory
    """
    return psutil.virtual_memory().available / div


MAXCHUNK_STRATEGY = max_available_div


class FDataObj:
    """ Wrapper for dataobj that returns floating point values from indexing.
    """

    # Signal that this is a proxy (for a proxy).
    is_proxy = True

    def __init__(self, dataobj, dtype=np.float64):
        dtype = np.dtype(dtype)
        if not issubclass(dtype.type, np.inexact):
            raise ValueError(f'{dtype} should be floating point type')
        self._dataobj = dataobj
        self.dtype = dtype
        self.shape = dataobj.shape
        self.ndim = dataobj.ndim
        self.order = getattr(dataobj, 'order', None)

    def __getitem__(self, slicer):
        """ Return image data as floating point type ``self.dtype``.
        """
        return np.asanyarray(self._dataobj[slicer], dtype=self.dtype)

    def chunk_sizes(self, maxchunk=None):
        """ Calculate chunk sizes for dataobj shape

        Parameters
        ----------
        maxchunk : None or int, optional
            The largest allowable chunk sizes in bytes.

        Returns
        -------
        chunk_sizes : list
            Chunk sizes for Dask array creation, being number of elements in
            one chunk over all axes of array in ``self.dataobj``.
        """
        return default_chunks(self, maxchunk)

    def reshape(self, shape):
        """ Return object reshaped to `shape`

        Notes
        -----
        There is no requirement that the contained ``dataobj`` has a
        ``reshape`` method; if not, calling this method will raise an
        ``AttributeError``.
        """
        return self.__class__(self._dataobj.reshape(shape), self.dtype)


def default_chunks(arr, order=None, maxchunk=None):
    """ Calculate chunk sizes for array-like `arr` from shape and `order`.

    Parameters
    ----------
    arr : array-like
        Object with attributes `.shape` and `.dtype`.  It can also have
        attribute `order`, and / or `.flags.f_contiguous`, used only if `order`
        below is None.
    order : {None, "F", "C"}
        Memory ordering of array.
    maxchunk : None or int, optional
        The largest allowable chunk sizes in bytes.

    Returns
    -------
    chunk_sizes : list
        Chunk sizes for Dask array creation, being number of elements in
        one chunk over all axes of array in ``self.dataobj``.
    """
    if order is None:  # FDataObj, dataobj
        order = getattr(arr, 'order', None)  # FDataObj, dataobj
    if order is None:  # ndarray
        order = getattr(getattr(arr, 'flags', None), 'f_contiguous', 'C')
    ndim = len(arr.shape)
    sizes = [None] * ndim
    if maxchunk is None:
        maxchunk = MAXCHUNK_STRATEGY()
    chunk_size = np.prod(arr.shape) * arr.dtype.itemsize
    if chunk_size <= maxchunk:
        return sizes
    axis_nos = range(ndim)
    if order == 'F':
        axis_nos = axis_nos[::-1]
    for axis_no in axis_nos:
        chunk_size //= arr.shape[axis_no]
        n_chunks = maxchunk // chunk_size
        if n_chunks:
            sizes[axis_no] = int(n_chunks)
            return sizes
        sizes[axis_no] = 1
    return sizes


_DIMNO2NAME = {
    None: None,
    0: 'i',
    1: 'j',
    2: 'k'}


_DIMNAME2NO = {v: k for k, v in _DIMNO2NAME.items()}


_TIME_UNIT_SCALER = {
    'sec': 1,
    'msec': 1 / 1000,
    'usec': 1 / 1_000_000}


class NiHeader2Attrs:

    def __init__(self, header):
        self.header = nib.Nifti1Header.from_header(header)

    def get_dim_labels(self):
        freq_dim, phase_dim, slice_dim = self.header.get_dim_info()
        return {'xib-FrequencyEncodingDirection': _DIMNO2NAME[freq_dim],
                'PhaseEncodingDirection': _DIMNO2NAME[phase_dim],
                'SliceEncodingDirection': _DIMNO2NAME[slice_dim]}

    def get_slice_timing(self):
        hdr = self.header
        freq_dim, phase_dim, slice_dim = hdr.get_dim_info()
        if not slice_dim:
            return None
        duration = hdr.get_slice_duration()
        if duration == 0:
            return None
        slice_start, slice_end = hdr['slice_start'], hdr['slice_end']
        n_slices = hdr.get_n_slices()
        if slice_start != 0 or slice_end != n_slices - 1:
            return None
        try:
            return list(hdr.get_slice_times())
        except HeaderDataError:
            return None

    def get_repetition_time(self):
        hdr = self.header
        zooms = hdr.get_zooms()
        if len(zooms) < 4:
            return None
        time_zoom = zooms[3]
        space_units, time_units = hdr.get_xyzt_units()
        if time_units == 'unknown':
            return None
        return _TIME_UNIT_SCALER[time_units] * time_zoom

    def get_affines(self):
        """ Collect valid affines
        """
        hdr = self.header
        affines = {}
        for affine_type in 'qform', 'sform':
            code = hdr.get_value_label(affine_type + '_code')
            if code != 'unknown':
                affine = getattr(hdr, 'get_' + affine_type)()
                affines[code] = affine.tolist()
        return affines

    def to_attrs(self):
        attrs = self.get_dim_labels()
        attrs['SliceTiming'] = self.get_slice_timing()
        attrs['RepetitionTime'] = self.get_repetition_time()
        attrs['xib-affines'] = self.get_affines()
        return {k: v for k, v in attrs.items() if v is not None}


class Attrs2NiHeader:
    """ Class writes BIDS attribute dictionary into NIfTI header.
    """

    def __init__(self, header=None, attrs=None):
        self.header = nib.Nifti1Header.from_header(header)
        self.attrs = {} if attrs is None else attrs.copy()

    def set_dim_labels(self):
        # Assume canonical axis order.
        d = self.attrs
        self.header.set_dim_info(
            freq=_DIMNAME2NO.get(d.get('xib-FrequencyEncodingDirection')),
            phase=_DIMNAME2NO.get(d.get('PhaseEncodingDirection')),
            slice=_DIMNAME2NO.get(d.get('SliceEncodingDirection')))

    def set_slice_timing(self):
        slice_times = self.attrs.get('SliceTiming')
        if slice_times is None:
            return
        try:
            return self.header.set_slice_times(slice_times)
        except HeaderDataError:
            pass

    def set_repetition_time(self):
        hdr = self.header
        zooms = np.array(hdr.get_zooms())
        # Assume header record of image already has correct shape.
        if len(zooms) < 4:
            return
        TR = self.attrs.get('RepetitionTime')
        TR, units = (0, None) if TR is None else (TR, 'sec')
        zooms[3] = TR
        hdr.set_xyzt_units(t=units)
        hdr.set_zooms(zooms)

    def set_affines(self):
        """ Set valid affines to header.
        """
        hdr = self.header
        affines = self.attrs.get('xib-affines', {})
        methods = [hdr.set_sform, hdr.set_qform]
        for code in ('mni', 'talairach', 'template', 'aligned', 'scanner'):
            if code in affines:
                methods.pop()(affines[code], code)
                if not methods:
                    break

    def updated_header(self):
        self.set_dim_labels()
        self.set_slice_timing()
        self.set_affines()
        self.set_repetition_time()
        return self.header


def hdr2attrs(header):
    # We could try extracting more information from other file types, but
    return NiHeader2Attrs(header).to_attrs()


def load_zarr(url_or_path, **kwargs):
    r""" Load image from Zarr/Xibabel-format data at `url_or_path`

    Parameters
    ----------
    url_or_path : str or Path
    \*\*kwargs : dict
        Any remaining named arguments passed to `xr.open_dataarray`.

    Returns
    -------
    ximg : Xibabel image

    Raises
    ------
    XibFileError
        If there is no file corresponding to `url_or_path`.
    """
    return xr.open_dataarray(url_or_path,
                             engine='zarr',
                             chunks='auto',
                             **kwargs)


class XibError(Exception):
    """ Errors from Xibabel file inference and operations
    """


class XibFormatError(XibError):
    """ Errors from Xibabel format specification or inference
    """


class XibFileError(XibError):
    """ Error from Xibabel file operations
    """


_JSON_MARKER = '__json__'


def _json_attrs2attrs(attrs):
    """ Read JSON formed for encoding nested structures to netCDF
    """
    out = {}
    for key, value in attrs.items():
        if (isinstance(value, list) and
            len(value) == 2 and
            value[0] == _JSON_MARKER):
            value = json.loads(value[1])
        out[key] = value
    return out


def _1d_arrayable(v):
    try:
        arr = np.array(v)
    except ValueError:
        return False
    return arr.ndim < 2


class NPEncoder(json.JSONEncoder):

    def default(self, obj):
        if hasattr(obj, 'tolist'):
            return obj.tolist()
        # No supported serialization, pass to default encoder to raise.
        return super().default(obj)


_jdumps = partial(json.dumps, cls=NPEncoder)


def _attrs2json_attrs(attrs):
    """ Write JSON formed for encoding nested structures to netCDF
    """
    out = {}
    for key, value in attrs.items():
        if (isinstance(value, dict) or
            (isinstance(value, (list, tuple)) and not _1d_arrayable(value))):
            value = [_JSON_MARKER, _jdumps(value)]
        out[key] = value
    return out


def _check_netcdf():
    if importlib.util.find_spec('h5netcdf') is None:
        raise XibFileError('Please install h5netcdf module to load netCDF')


def load_netcdf(url_or_path, **kwargs):
    r""" Load image from netCDF/Xibabel-format data at `url_or_path`

    Parameters
    ----------
    url_or_path : str or Path
    \*\*kwargs : dict
        Any remaining named arguments passed to `xr.open_dataarray`.

    Returns
    -------
    ximg : Xibabel image

    Raises
    ------
    XibFileError
        If there is no file corresponding to `url_or_path`.
    """
    _check_netcdf()
    with fsspec.open(url_or_path) as fobj:
        img = xr.open_dataarray(fobj, engine='h5netcdf', **kwargs)
    img.attrs = _json_attrs2attrs(img.attrs)
    return img


def load(url_or_path, *, format=None, **kwargs):
    r""" Load image Xibabel or Nibbel image at `url_or_path`

    Parameters
    ----------
    url_or_path : str or Path
    format : None or str, optional
        If None, infer format from `url_or_path`, otherwise one of
        {', '.join(f'{e}' for e in PROCESSORS.format_processors)}
    \*\*kwargs : dict
        Any remaining named arguments passed to backend corresponding to
        `format`.

    Returns
    -------
    ximg : Xibabel image

    Raises
    ------
    XibFileError
        If there is no file corresponding to `url_or_path`.
    """
    if format is None:
        format = PROCESSORS.guess_format(url_or_path)
    return PROCESSORS.get_loader(format)(url_or_path, **kwargs)


# Extensions we will search for valid images paired with JSON files.
_JSON_PAIRED_EXTS = ('.nii.gz', '.nii')


def drop_suffix(in_path, suffix):
    """ Drop suffix in `suffixes` from str or ``Path`` `in_path`

    `suffix` can be ``str`` (suffix to drop) or sequence.  If sequence, drops
    first matching suffix in sequence `suffix`.

    Parameters
    ----------
    in_path : str or Path
        Path from which to drop suffix.
    suffix : str or sequence
        If ``str``, suffix to drop.  If sequence, search for each suffix, and
        drop first found suffix from `in_path`.

    Returns
    -------
    out_path : str or Path
        Return type matches that of `in_path`.
    """
    if not hasattr(in_path, 'is_file'):
        return _drop_suffix_str(in_path, suffix)
    return in_path.with_name(_drop_suffix_str(in_path.name, suffix))


def _drop_suffix_str(path_str, suffix):
    suffixes = (suffix,) if isinstance(suffix, str) else suffix
    for suffix in suffixes:
        if path_str.endswith(suffix):
            return path_str[:-(len(suffix))]
    return path_str


def replace_suffix(in_path, old_suffix, new_suffix):
    """ Replace `in_path` suffix with `new_suffix`, allowing for `old_suffix`

    Always replace suffix of `in_path`, if present, but allowing for any
    special (multi-dot) suffixes in `old_sequence`.

    `old_suffix` can be ``str`` (suffix to drop) or sequence.  If sequence,
    drops first matching suffix in sequence `suffixes` before appending
    `new_suffix`.

    Parameters
    ----------
    in_path : str or Path
        Path from which to replace suffix.
    old_suffix : str or sequence
        If ``str``, suffix to drop before replacing.  If sequence, search for
        each suffix, and drop first found suffix from `in_path`.
    new_suffix : str
        Suffix to append.

    Returns
    -------
    out_path : str or Path
        Return type matches that of `in_path`.
    """
    is_path = hasattr(in_path, 'is_file')
    path_str = in_path.name if is_path else in_path
    dropped = _drop_suffix_str(path_str, old_suffix)
    replaced = op.splitext(dropped)[0] + new_suffix
    return in_path.with_name(replaced) if is_path else replaced


def _valid_or_raise(fs, url_base, exts):
    for ext in exts:
        target_url = replace_suffix(url_base, (), ext)
        if fs.exists(target_url):
            return fsspec.open(target_url, compression='infer')
    msg_suffix = (('one of ' if len(exts) > 1 else '') +
                  ', '.join(f"'{e}'" for e in exts))
    raise XibFileError(
        f"No valid file matching '{url_base}' + {msg_suffix}")


def load_bids(url_or_path, *, require_json=True, **kwargs):
    r""" Load image from BIDS-format data at `url_or_path`

    `url_or_path` may point directly to ``.json`` file, or to Nibabel format
    file, for which we expect a ``.json`` file to be present.

    In the second case, if `require_json` is True, and we cannot find a
    matching `.json` file, raise error, otherwise return image as best read.

    Parameters
    ----------
    url_or_path : str or Path
    require_json : {True, False}, optional, keyword-only
        If True, raise error if `url_or_path` is an image, and there is no
        matching JSON file.
    \*\*kwargs : dict
        Any remaining named arguments passed to `fsspec.open`.

    Returns
    -------
    bids_ximg : Xibabel image

    Raises
    ------
    XibFileError
        If `require_json` is True, `url_or_path` does not name a ``.json`` file
        and there is no ``.json`` file corresponding to `url_or_path`.
    """
    sidecar_file = None
    # If url_or_path has .json suffix, search for matching image file.
    if str(url_or_path).endswith('.json'):
        sidecar_file = fsspec.open(url_or_path, **kwargs)
        fs = sidecar_file.fs
        if not fs.exists(url_or_path):
            raise XibFileError(
                f'{url_or_path} does not appear to exist')
        img_file = _valid_or_raise(fs,
                                   drop_suffix(url_or_path, '.json'),
                                   _JSON_PAIRED_EXTS)
    else:  # Image file extensions.  Search for JSON.
        img_file = fsspec.open(url_or_path, compression='infer', **kwargs)
        fs = img_file.fs
        if not fs.exists(url_or_path):
            raise XibFileError(
                f'{url_or_path} does not appear to exist')
        url_uncomp = drop_suffix(url_or_path, _comp_exts())
        url_json = replace_suffix(url_uncomp, (), '.json')
        if fs.exists(url_json):
            sidecar_file = fs.open(url_json)
        elif require_json:
            raise XibFileError(
                f'BIDS loading {url_or_path} but no corresponding '
                f'{url_json} file, and `require_json` is True')
    nib_img, attrs = _nibabel2img_attrs(img_file)
    if sidecar_file:
        with sidecar_file as fobj:
            attrs.update(json.load(fobj))
    return _img_attrs2ximg(
        nib_img,
        **{'attrs': attrs, 'name': _url2name(url_or_path)})


def load_nibabel(url_or_path, **kwargs):
    r""" Load image from Nibabel-format data at `url_or_path`

    `url_or_path` may point directly to ``.json`` file, or to Nibabel format
    file, for expect a ``.json`` file to be present, in which case we will load
    it.

    Parameters
    ----------
    url_or_path : str or Path
    \*\*kwargs : dict
        Any remaining named arguments passed to `fsspec.open`.

    Returns
    -------
    ximg : Xibabel image

    Raises
    ------
    XibFileError
        If there is no file corresponding to `url_or_path`.
    """
    return load_bids(url_or_path, require_json=False, **kwargs)


def from_array(arr, *, fa_kwargs=None, **kwargs):
    r""" Create image from array-like `arr`

    Assume array dimensions correspond to NIfTI labels.

    Parameters
    ----------
    arr : array-like
    chunks : 'auto' or int or tuple
        See documentation for ``dask.array.from_array``
    fa_kwargs : None or mapping, optional, keyword only
        Keyword arguments that we pass through the ``dask.array.from_array``
        function.
    \*\*kwargs : mapping, keyword arguments
        Keyword arguments to pass to ``DataArray`` constructor.  A common
        example is ``dims`` (None or sequence of dimension names).  If None,
        use NIfTI convention for coordinate names.  If sequence, it must match
        the number of dimensions of `arr`.

    Returns
    -------
    ximg : Xibabel image
    """
    arr_like = FDataObj(arr) if nib.is_proxy(arr) else arr
    ximg = _arr_attrs2ximg(arr_like, fa_kwargs=fa_kwargs, **kwargs)
    if 'dims' not in kwargs:  # Use NIfTI conventions
        ximg = _squeeze_time(ximg)
    return ximg


def _comp_exts():
    return tuple(f'.{ext}' for ext in fsspec.utils.compressions)


def _path2class(filename):
    compression_exts = _comp_exts()
    for klass in nib.all_image_classes:
        if len(klass.files_types) > 1:
            continue
        base, ext, gzext, ftype = parse_filename(filename,
                                                 klass.files_types,
                                                 compression_exts)
        if ftype == 'image':
            return klass
    raise XibFileError(f'No single-file Nibabel class for {filename}')


class FSFileHolder(FileHolder):

    def __del__(self):
        self.fileobj.close()


_NI_DIM_NAMES = ('i', 'j', 'k', 'time', 'p', 'q', 'r')
_NI_SPACE_DIMS = _NI_DIM_NAMES[:3]
_NI_TIME_DIM = _NI_DIM_NAMES[3]


def _nibabel2img_attrs(img_file):
    # Identify relevant files from img_file
    # Make file_map with opened files.
    if 'local' in img_file.fs.protocol:
        img = nib.load(img_file.path)
    else:  # Not local - use stream interface.
        img_klass = _path2class(img_file.path)
        # We are passing out opened fsspec files.
        fh = FileHolder(img_file.path, img_file.open())
        img = img_klass.from_file_map({'image': fh})
    return img, hdr2attrs(img.header)


def _img_attrs2ximg(obj, *, fa_kwargs=None, **kwargs):
    arr_like = FDataObj(obj.dataobj)
    fa_kwargs = {} if fa_kwargs is None else {}
    if not fa_kwargs.get('fancy'):  # Assume no fancy indexing by default.
        fa_kwargs['fancy'] = False
    if not fa_kwargs.get('chunks'):
        fa_kwargs['chunks'] = arr_like.chunk_sizes()
    ximg = _arr_attrs2ximg(arr_like,
                           fa_kwargs=fa_kwargs,
                           **kwargs)
    return _squeeze_time(ximg) if hasattr(obj, 'get_sform') else ximg


def _squeeze_time(ximg):
    # https://nifti.nimh.nih.gov/nifti-1/documentation/nifti1fields/nifti1fields_pages/dim.html
    # > If dim[4]=1 or dim[0] < 4, there is no time axis.
    if 'time' in ximg.dims and len(ximg['time']) == 1:
        ximg = ximg.sel(time=0)
    return ximg


def _arr_attrs2ximg(arr, *, fa_kwargs=None, **kwargs):
    fa_kwargs = {} if fa_kwargs is None else fa_kwargs
    if not fa_kwargs.get('chunks'):
        fa_kwargs['chunks'] = default_chunks(arr)
    shape = arr.shape
    if not kwargs.get('dims'):
        kwargs['dims'] = list(_NI_DIM_NAMES)[:len(shape)]
    kwargs['coords'] = _filled_coords(kwargs.get('coords', {}),
                                      kwargs['dims'],
                                      shape,
                                      kwargs.get('attrs', {}))
    return xr.DataArray(da.from_array(arr, **fa_kwargs), **kwargs)


def _filled_coords(coords, dims, shape, attrs):
    # Fill space coordinate axes
    out_coords = coords.copy()
    for ax_no, ax_name in enumerate(dims):
        if ax_name not in _NI_SPACE_DIMS or ax_name in out_coords:
            continue
        out_coords[ax_name] = xr.DataArray(
            np.arange(shape[ax_no]),
            dims=[ax_name])
    if _NI_TIME_DIM not in dims:
        return out_coords
    # Consider special cases for: hz, ppm, rad
    # https://nifti.nimh.nih.gov/nifti-1/documentation/nifti1fields/nifti1fields_pages/xyzt_units.html
    # Maybe set attrs['xib-time-unts'] from header, and use here.
    TR = attrs.get("RepetitionTime")
    if not TR:
        return out_coords
    # Add time axis coordinates.
    n_time = shape[dims.index(_NI_TIME_DIM)]
    out_coords[_NI_TIME_DIM] = xr.DataArray(
        np.arange(0, n_time) * TR,
        dims=[_NI_TIME_DIM],
        attrs={"units": "s"})
    return out_coords


def _url2name(url_or_path):
    name = drop_suffix(url_or_path, _comp_exts())
    return Path(name).stem


def save(obj, url_or_path, format=None, **kwargs):
    r""" Save Xibabel image at `url_or_path`

    Parameters
    ----------
    url_or_path : str or Path
    format : None or str, optional
        If None, infer format from `url_or_path`, otherwise one of
        'zarr', 'netcdf', 'bids', 'nibabel'.
    \*\*kwargs : dict
        Any remaining named arguments passed to backend corresponding to
        `format`.

    Returns
    -------
    ret : object
        Return depends on `format`.  See `save_zarr` and `save_netcdf` for
        examples.
    """
    if format is None:
        format = PROCESSORS.guess_format(url_or_path)
    return PROCESSORS.get_saver(format)(obj, url_or_path, **kwargs)


def save_zarr(obj, url_or_path, **kwargs):
    r""" Save Xibabel image at `url_or_path` in Zarr / Xibabel format.

    Parameters
    ----------
    url_or_path : str or Path
    \*\*kwargs : dict
        Arguments to pass to `to_zarr` method of `obj`.

    Returns
    -------
    zbe : ZarrBackendStore
    """
    return obj.to_zarr(url_or_path, mode='w', **kwargs)


def save_netcdf(obj, url_or_path, **kwargs):
    r""" Save Xibabel image at `url_or_path` in netCDF / Xibabel format.

    Parameters
    ----------
    url_or_path : str or Path
    \*\*kwargs : dict
        Arguments to pass to `to_netcdf` method of `obj`.

    Returns
    -------
    None
    """
    _check_netcdf()
    out = obj.copy()  # Shallow copy by default.
    out.attrs = _attrs2json_attrs(out.attrs)
    return out.to_netcdf(url_or_path, engine='h5netcdf', **kwargs)


def save_bids(obj, url_or_path, **kwargs):
    r""" Save Xibabel image at `url_or_path` as NIfTI / BIDS JSON

    Parameters
    ----------
    url_or_path : str or Path
    \*\*kwargs : dict
        Any remaining named arguments passed to `fsspec.open`.

    Returns
    -------
    None
    """
    nib_img, attrs = to_nifti(obj)
    if str(url_or_path).endswith('.json'):
        json_uop = url_or_path
        img_uop = replace_suffix(url_or_path, '.json', '.nii.gz')
    else:
        img_uop = url_or_path
        json_uop = replace_suffix(
            drop_suffix(url_or_path, _comp_exts()),
            (),
            '.json')
    img_file = fsspec.open(img_uop, compression='infer', **kwargs)
    sidecar_kwargs = merge(kwargs, {'mode': 'wt'})
    sidecar_file = img_file.fs.open(json_uop, **sidecar_kwargs)
    with sidecar_file as f:
        f.write(_jdumps(attrs))
    if 'local' in img_file.fs.protocol:
        nib.save(nib_img, img_file.path)
        return
    # Not local - use stream interface.
    img_klass = _path2class(img_uop.path)
    # We are passing out opened fsspec files.
    with img_uop as fh:
        img_klass.to_file_map({'image': fh})


class Processors:

    format_processors = {
        'zarr': dict(exts=('ximg',),
                     loader=load_zarr,
                     saver=save_zarr),
        'netcdf': dict(exts=('nc',),
                     loader=load_netcdf,
                     saver=save_netcdf),
        'bids': dict(exts=('json',),
                     loader=load_bids,
                     saver=save_bids),
        'nibabel': dict(exts=(),  # Defer to Nibabel for extensions
                     loader=load_nibabel,
                     saver=save_bids),
    }

    def __init__(self):
        self.loaders = set()
        self.savers = set()
        self.ext2fmt = {}
        for fmt, info in self.format_processors.items():
            if info['loader']:
                self.loaders.add(fmt)
            if info['saver']:
                self.savers.add(fmt)
            for ext in info['exts']:
                self.ext2fmt[ext] = fmt

    def get_loader(self, fmt):
        return self.get_processor(fmt, 'loader')

    def get_saver(self, fmt):
        return self.get_processor(fmt, 'saver')

    def get_processor(self, fmt, ptype='loader'):
        fmt = 'nibabel' if fmt is None else fmt
        pset = (self.loaders if ptype == 'loader'
                else self.savers)
        if fmt not in pset:
            raise XibFormatError(
                f"Cannot use format '{fmt}' as {ptype} for image; "
                f"valid formats are {','.join(pset)}")
        return self.format_processors[fmt][ptype]

    def guess_format(self, file_path):
        suff = str(file_path).split('.')[-1]
        return self.ext2fmt.get(suff)


PROCESSORS = Processors()


def _ni_sort_expand_dims(img_dims):
    # Allow for 3D images
    target_dims = _NI_SPACE_DIMS
    if not set(img_dims).issubset(_NI_SPACE_DIMS):
        target_dims += (_NI_TIME_DIM,)
    x_dims = []
    x_axes = []
    out_order = []
    for expected_pos, expected_dim in enumerate(target_dims):
        try:
            curr_pos = img_dims.index(expected_dim)
        except ValueError:
            x_dims.append(expected_dim)
            x_axes.append(expected_pos)
            continue
        out_order.append(img_dims[curr_pos])
    return (out_order + [d for d in img_dims if d not in out_order],
            x_dims,
            x_axes)


def to_nifti(ximg):
    # Reorient, expand missing dimensions
    order, dims, axes = _ni_sort_expand_dims(ximg.dims)
    ximg = ximg.transpose(*order).expand_dims(dims, axes)
    # Adjust affines to current state of ximg.
    back = ximg.xi.with_updated_affines()
    # Build preliminary header.
    hdr = nib.Nifti1Header()
    hdr.set_data_shape(back.shape)
    # Build header from attributes.
    hdr = Attrs2NiHeader(hdr, back.attrs).updated_header()
    return nib.Nifti1Image(back, None, hdr), back.attrs


@xr.register_dataarray_accessor("xi")
class XiAccessor:

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def get_affines(self):
        """ Get spatial affines from attributes of image

        Returns
        -------
        affines : dict
            Dictionary with key, value pairs where key is string giving space
            to which affine maps, and values are 4x4 spatial affine arrays.

        Notes
        -----
        Returned `affines` are copies; if you modify the arrays in `affines`,
        this will not affect the original image.
        """
        aff_d = self._obj.attrs.get('xib-affines', {})
        return {space: np.array(aff) for space, aff in aff_d.items()}

    def set_affines(self, affines):
        """ Set spatial affines to attributes of image

        Affines overwrite existing affines.  If the affine does not exist, we
        append to the affines dictionary.

        Parameters
        ----------
        affines : dict
            Dictionary where key, value pairs are affine space name and 4x4
            affine array, respectively.

        Notes
        -----
        Set `affines` are copies; if you modify the arrays in `affines`,
        this will not affect the image affines.  The set affines overwrite
        correspnding existing affines.
        """
        aff_d = self._obj.attrs.get('xib-affines', {})
        for space, aff in affines.items():
            aff_d[space] = np.array(aff)
        self._obj.attrs['xib-affines'] = aff_d

    def with_updated_affines(self):
        """ Return image with affines, coordinates updated to match state.

        Return image `adj_img` where the affines and coordinates correctly
        reflect any detected slicing of the original image.

        Returns
        -------
        adj_img : Xibabel image
            Image with adjusted affines, and for which spatial coordinate
            indices have been reset to sequential 0-based default.

        Notes
        -----
        The affines and i, j, k coordinates reflect the state of the image as
        of the last call to this function, or as of image load.

        The affines are always 4x4, and always refer to i, j, k space,
        regardless of dimension ordering, and of the presence of i, j, k
        dimensions.
        """
        adj_img = self._obj.copy()
        assert adj_img.attrs is not self._obj.attrs
        out_affines = {}
        for space, affine in self.get_affines().items():
            out_affines[space] = self._adjusted_affine(affine, adj_img.coords)
        adj_img.attrs['xib-affines'] = out_affines
        return adj_img.xi._with_reset_coordinates()

    def _adjusted_affine(self, affine, coords):
        vox_origin = np.zeros(3)
        vox_scalings = np.ones(3)
        for i, name in enumerate(_NI_SPACE_DIMS):
            vox_indices = np.ravel(coords.get(name, 0))
            vox_origin[i] = vox_indices[0]
            if len(vox_indices) > 1:
                vd = np.diff(vox_indices)
                if any(vd[1:] != vd[0]):
                    raise XibFormatError(
                        f'Cannot handle irregular voxel spacing for "{name}"')
                vox_scalings[i] = vd[0]
        return from_matvec(affine[:3, :3] * vox_scalings,
                           apply_affine(affine, vox_origin))

    def _with_reset_coordinates(self):
        ximg = self._obj
        coords = {}
        for name in _NI_SPACE_DIMS:
            if name not in ximg.dims:
                coords[name] = xr.DataArray(0)
                continue
            coords[name] = xr.DataArray(
                np.arange(ximg[name].shape[0]),
                dims=[name])
        return ximg.assign_coords(coords)



""" Test markers
"""

from importlib.util import find_spec

import pytest

nipy_test = pytest.mark.skipif(not find_spec('nipy'),
                               reason="Need Nipy for this test")

h5netcdf_test = pytest.mark.skipif(not find_spec('h5netcdf'),
                                   reason='Need h5netcdf module for test')



""" Testing utilties and data

See :mod:`fetcher` for the data fetcher, and instructions for adding files.
"""

from collections.abc import Mapping
from pathlib import Path

import numpy as np

from ..loaders import drop_suffix
from .fetcher import Fetcher


fetcher = Fetcher(config_path=Path(__file__).parent)
get_set = fetcher.get_set
get_file = fetcher.get_file
have_file = fetcher.have_file
DATA_PATH = fetcher.data_path
skip_without_file = fetcher.skip_without_file


# Example files
JC_EG_FUNC = (DATA_PATH / 'ds000009' / 'sub-07' / 'func' /
              'sub-07_task-balloonanalogrisktask_bold.nii.gz')
JC_EG_FUNC_JSON = drop_suffix(JC_EG_FUNC, ('.nii.gz',)).with_suffix('.json')
JC_EG_ANAT= (DATA_PATH / 'ds000009' / 'sub-07' / 'anat' /
             'sub-07_T1w.nii.gz')
JC_EG_ANAT_JSON = drop_suffix(JC_EG_ANAT, ('.nii.gz',)).with_suffix('.json')
JH_EG_FUNC = (DATA_PATH / 'ds000105' / 'sub-1' / 'func' /
              'sub-1_task-objectviewing_run-01_bold.nii.gz')


def _dts_st(dts, dtt, op=all):
    return op(np.issubdtype(dt, dtt) for dt in dts)


def arr_dict_allclose(dict1, dict2, *args, **kwargs):
    r""" True if two dicts are equal or allclose, where dicts may contain arrays

    Parameters
    ----------
    dict1 : object
        object (usually dictionary).
    dict2 : object
        object (usually dictionary).
    \*args : sequence
        Positional arguments to pass to allclose.
    \**args : sequence
        Keyword arguments to pass to allclose.

    Returns
    -------
    isclose : bool
        True if all elements in `dict2` are equal to or np.allclose to
        equivalent elements in `dict1`.
    """
    if not (isinstance(dict1, Mapping) and isinstance(dict2, Mapping)):
        return False
    if not set(dict1) == set(dict2):
        return False
    for key, value1 in dict1.items():
        value2 = dict2[key]
        if isinstance(value1, Mapping):
            if not isinstance(value2, Mapping):
                return False
            if not arr_dict_allclose(value1, value2):
                return False
            continue
        v1_arr = np.array(value1)
        v2_arr = np.array(value2)
        dts = v1_arr.dtype, v2_arr.dtype
        if _dts_st(dts, np.inexact, any) and _dts_st(dts, np.number, all):
            if not np.allclose(v1_arr, v2_arr, *args, **kwargs):
                return False
        elif isinstance(value1, np.ndarray) or isinstance(value2, np.ndarray):
            if np.any(v1_arr != value2):
                return False
        elif value1 != value2:
            return False
    return True



""" Xibabel neuroimaging interface

Xibabel is an experimental package for working with neuroimaging data formats.
"""

from .loaders import load, save, from_array, to_nifti  # noqa:F401

__version__ = "0.0.1rc1"



""" Test loaders
"""

from pathlib import Path
from copy import deepcopy
import os
from itertools import product, permutations
import json
import shutil

import numpy as np

import nibabel as nib

import xarray as xr

from xibabel import loaders
from xibabel.loaders import (FDataObj, load_bids, load_nibabel, load, save,
                             PROCESSORS, _json_attrs2attrs, drop_suffix,
                             replace_suffix, _attrs2json_attrs, hdr2attrs,
                             _path2class, XibFileError, XibFormatError,
                             to_nifti, _ni_sort_expand_dims, _NI_SPACE_DIMS,
                             _NI_TIME_DIM, _jdumps, from_array)
import nibabel.testing as nit
from xibabel.xutils import merge
from xibabel.testing import (JC_EG_FUNC, JC_EG_FUNC_JSON, JC_EG_ANAT,
                             JC_EG_ANAT_JSON, JH_EG_FUNC, skip_without_file,
                             fetcher, arr_dict_allclose)
from xibabel.tests.markers import h5netcdf_test

import pytest

rng = np.random.default_rng()


class FakeProxy:

    def __init__(self, array, order=None):
        self._array = array
        self.order = order
        self.ndim = array.ndim
        self.shape = array.shape

    def __getitem__(self, slicer):
        return self._array[slicer]

    def __array__(self):
        return self[:]

    def reshape(self, shape):
        return self._array.reshape(shape)


def test_fdataobj_basic():
    arr = np.arange(24).reshape((2, 3, 4))
    proxy = FakeProxy(arr)
    fproxy = FDataObj(proxy)
    assert fproxy.ndim == arr.ndim
    assert fproxy.shape == arr.shape
    assert fproxy.order is None
    assert np.all(fproxy[1, 2, :] == arr[1, 2, :])
    assert arr[1, 2, :].dtype == np.arange(2).dtype
    assert fproxy.dtype == np.dtype(np.float64)
    assert fproxy[1, 2, :].dtype == np.dtype(np.float64)
    fproxy = FDataObj(proxy, dtype=np.float32)
    assert fproxy.dtype == np.dtype(np.float32)
    with pytest.raises(ValueError, match='should be floating point type'):
        FDataObj(proxy, dtype=int)
    rprox = fproxy.reshape((6, 4))
    assert rprox.shape == (6, 4)
    assert np.all(rprox[:] == arr.reshape((6, 4)))


def test_chunking(monkeypatch):
    arr_shape = 10, 20, 30
    arr = rng.normal(size=arr_shape)
    fproxy = FDataObj(FakeProxy(arr))
    c_fproxy = FDataObj(FakeProxy(arr, order='C'))
    f_fproxy = FDataObj(FakeProxy(arr, order='F'))
    f64s = np.dtype(float).itemsize
    N = None
    for strategy, c_exp_sizes, f_exp_sizes in (
        (lambda : arr.size * f64s, [N, N, N], [N, N, N]),
        (lambda : arr.size * f64s - 1, [9, N, N], [N, N, 29]),
        (lambda : arr.size * f64s / 2 - 1, [4, N, N], [N, N, 14]),
        (lambda : arr.size * f64s / 10 - 1, [1, 19, N], [N, N, 2]),
        (lambda : arr.size * f64s / 30 - 1, [1, 6, N], [N, 19, 1]),
        (lambda : 11 * f64s, [1, 1, 11], [None, 1, 1]),
        (lambda : 10 * f64s, [1, 1, 10], [None, 1, 1]),
        (lambda : 2 * f64s, [1, 1, 2], [2, 1, 1]),
    ):
        monkeypatch.setattr(loaders, "MAXCHUNK_STRATEGY", strategy)
        assert fproxy.chunk_sizes() == c_exp_sizes
        assert c_fproxy.chunk_sizes() == c_exp_sizes
        assert f_fproxy.chunk_sizes() == f_exp_sizes


def out_back(img, out_path):
    if out_path.is_file():
        os.unlink(out_path)
    nib.save(img, out_path)
    img = nib.load(out_path)
    return img, hdr2attrs(img.header)


def out_back_xi(ximg, out_path):
    if out_path.is_file():
        os.unlink(out_path)
    save(ximg, out_path)
    # Compute on file to allow for later deletion of source file.
    return load(out_path).compute()


def test_nibabel_tr(tmp_path):
    # Default image.
    arr = np.zeros((2, 3, 4))
    img = nib.Nifti1Image(arr, np.eye(4), None)
    out_path = tmp_path / 'test.nii'
    back_img, attrs = out_back(img, out_path)
    exp_attrs = {'xib-affines': {'aligned': np.eye(4).tolist()}}
    assert attrs == exp_attrs
    arr = np.zeros((2, 3, 4, 6))
    img = nib.Nifti1Image(arr, np.eye(4), None)
    back_img, attrs = out_back(img, out_path)
    assert attrs == exp_attrs
    img.header.set_xyzt_units('mm', 'sec')
    back_img, attrs = out_back(img, out_path)
    exp_attrs.update({'RepetitionTime': 1.0})
    assert attrs == exp_attrs
    img.header.set_xyzt_units('mm', 'msec')
    back_img, attrs = out_back(img, out_path)
    assert attrs['RepetitionTime'] == 1 / 1000
    img.header.set_xyzt_units('mm', 'usec')
    back_img, attrs = out_back(img, out_path)
    assert attrs['RepetitionTime'] == 1 / 1_000_000


def test_nibabel_slice_timing(tmp_path):
    # Default image.
    arr = np.zeros((2, 3, 4, 5))
    img = nib.Nifti1Image(arr, np.eye(4), None)
    nib_path = tmp_path / 'test.nii'
    back_img, attrs = out_back(img, nib_path)
    # Check attrsdata.
    exp_attrs = {'xib-affines': {'aligned': np.eye(4).tolist()}}
    assert attrs == exp_attrs
    # Load ximg for comparison.
    ximg = load(nib_path).compute()  # Get data from file.
    assert ximg.attrs == exp_attrs
    # Try setting dimension information.
    img.header.set_dim_info(None, None, 1)
    nib_path2 = tmp_path / 'test2.nii'
    back_img, attrs = out_back(img, nib_path2)
    assert attrs == merge(exp_attrs, {'SliceEncodingDirection': 'j'})
    img.header.set_dim_info(1, 0, 2)
    back_img, attrs = out_back(img, nib_path2)
    exp_dim = merge(exp_attrs,
                    {'PhaseEncodingDirection': 'i',
                     'xib-FrequencyEncodingDirection': 'j',
                     'SliceEncodingDirection': 'k'})
    assert attrs == exp_dim
    img.header.set_slice_duration(1 / 4)
    back_img, attrs = out_back(img, nib_path2)
    assert attrs == exp_dim
    img.header['slice_start'] = 0
    back_img, attrs = out_back(img, nib_path2)
    assert attrs == exp_dim
    img.header['slice_end'] = 3
    back_img, attrs = out_back(img, nib_path2)
    assert attrs == exp_dim
    # Has time dimension.
    assert ximg.dims == tuple('ijk') + ('time',)
    # Check setting slice timing.
    img.header['slice_code'] = 4  # NIFTI_SLICE_ALT_DEC
    # This fills in the times.
    back_img, attrs = out_back(img, nib_path2)
    exp_timed = exp_dim.copy()
    slice_times = [0.75, 0.25, 0.5, 0]
    exp_timed['SliceTiming'] = slice_times
    assert attrs == exp_timed
    # Reset image back to default.
    back_ximg = out_back_xi(ximg, nib_path2)
    assert ximg.attrs == exp_attrs
    # Use header stuff to set slice timing.
    ximg.attrs['SliceTiming'] = slice_times
    back_ximg = out_back_xi(ximg, nib_path2)
    assert np.allclose(back_ximg.attrs['SliceTiming'], slice_times)


def _arr2ximg(arr, out_path):
    img = nib.Nifti1Image(arr, np.eye(4), None)
    nib.save(img, out_path)
    return load(out_path).compute()  # The file may later be deleted.


def test_nifti_load_save(tmp_path):
    # Default image.
    shape = (2, 3, 4, 5)
    arr = np.arange(np.prod(shape), dtype=float).reshape(shape)
    out_path = tmp_path / 'test.nii'
    ximg = _arr2ximg(arr, out_path)
    assert np.allclose(ximg, arr)
    assert ximg.dims == ('i', 'j', 'k', 'time')
    assert ximg.attrs.get('RepetitionTime') is None
    shape = (2, 3, 4, 5, 1)
    arr = np.arange(np.prod(shape), dtype=float).reshape(shape)
    out_path = tmp_path / 'test2.nii'
    ximg = _arr2ximg(arr, out_path)
    assert np.allclose(ximg, arr)
    assert ximg.dims == ('i', 'j', 'k', 'time', 'p')
    ximg.attrs['RepetitionTime'] = 2.0
    back_ximg = out_back_xi(ximg, tmp_path / 'test2a.nii')
    assert back_ximg.dims == ('i', 'j', 'k', 'time', 'p')
    assert back_ximg.attrs['RepetitionTime'] == 2.0
    shape = (2, 3, 4, 1, 5, 1)
    arr = np.arange(np.prod(shape), dtype=float).reshape(shape)
    out_path = tmp_path / 'test3.nii'
    ximg = _arr2ximg(arr, out_path)
    assert np.allclose(ximg, np.reshape(arr, (2, 3, 4, 5, 1)))
    assert ximg.dims == ('i', 'j', 'k', 'p', 'q')


@pytest.mark.parametrize("filename",
                         ('minc1_4d.mnc',
                          'minc2_4d.mnc',
                          'example_nifti2.nii.gz',
                          'phantom_varscale.PAR',
                         ))
def test_other_load(filename, tmp_path):
    in_path = nit.get_test_data() / filename
    nib_img = nib.load(in_path)
    ximg = load(in_path)
    assert nib_img.shape == ximg.shape
    assert len(ximg.dims) == len(nib_img.shape)
    assert np.allclose(nib_img.get_fdata(), ximg)


def test_from_array():
    # Without axis names, uses NIfTI convention
    a2d = rng.normal(size=(3, 4))
    ximg = from_array(a2d)
    assert ximg.dims == tuple('ij')
    assert np.all(a2d == ximg)
    a3d = rng.normal(size=(3, 4, 5))
    ximg = from_array(a3d)
    assert ximg.dims == tuple('ijk')
    assert np.all(a3d == ximg)
    a4d = rng.normal(size=(3, 4, 5, 6))
    ximg = from_array(a4d)
    assert ximg.dims == tuple('ijk') + ('time',)
    assert np.all(a4d == ximg)
    s5d = (3, 4, 5, 1, 6)
    a5d = rng.normal(size=s5d)
    ximg = from_array(a5d)
    assert ximg.dims == tuple('ijkp')
    assert np.all(a5d.reshape((3, 4, 5, 6)) == ximg)
    # With axis names, follows axis names.
    names = tuple('ji')
    ximg = from_array(a2d, dims=names)
    assert ximg.dims == names
    assert np.all(a2d == ximg)
    assert tuple(ximg.coords) == names
    names = tuple('pki')
    ximg = from_array(a3d, dims=names)
    assert ximg.dims == names
    # Only spatial axes have default coordinates.
    assert sorted(ximg.coords) == ['i', 'k']
    assert np.all(a3d == ximg)
    names = ('j', 'time', 'k', 'i')
    ximg = from_array(a4d, dims=names)
    assert ximg.dims == names
    assert sorted(ximg.coords) == ['i', 'j', 'k']
    assert np.all(a4d == ximg)
    # 5D with length 1 dimension not dropped by default.
    names = ('time',) + tuple('ikjn')
    ximg = from_array(a5d, dims=names)
    assert ximg.dims == names
    assert sorted(ximg.coords) == list('ijk')
    assert np.all(a5d == ximg)
    # With repetition time, we get a time axis
    attrs = {'RepetitionTime': 2.5}
    ximg = from_array(a5d, dims=names, attrs=attrs)
    assert ximg.dims == names
    assert sorted(ximg.coords) == ['i', 'j', 'k', 'time']
    assert np.all(a5d == ximg)
    assert np.all(ximg.coords['i'] == np.arange(4))
    assert np.all(ximg.coords['k'] == np.arange(5))
    # If you specify coordinates, we prefer the specified coordinates.
    coords = {'i': xr.DataArray(np.arange(2, 6), dims=['i'])}
    ximg = from_array(a5d, dims=names, attrs=attrs, coords=coords)
    assert np.all(ximg.coords['i'] == np.arange(2, 6))
    assert np.all(ximg.coords['k'] == np.arange(5))


def test_load_save_arr(tmp_path):
    out_path = tmp_path / 'test.nii'
    a3d = rng.normal(size=(3, 4, 5))
    ximg = from_array(a3d)
    assert ximg.xi.get_affines() == {}
    save(ximg, out_path)
    ximg_back = load(out_path)
    assert ximg_back.shape == (3, 4, 5)
    assert ximg_back.xi.get_affines() == {}


def test_get_set_affines():
    a3d = rng.normal(size=(3, 4, 5))
    ximg = from_array(a3d)
    assert ximg.dims == tuple('ijk')
    assert ximg.xi.get_affines() == {}
    new_affine = np.diag([1.2, 2.3, 3.4, 1])
    ximg.xi.set_affines({'scanner': new_affine})
    assert arr_dict_allclose(ximg.xi.get_affines(), {'scanner': new_affine})
    new_affine2 = new_affine.copy()
    new_affine2[:3, 3] = 10, 11, 12
    ximg.xi.set_affines({'scanner': new_affine2})
    assert arr_dict_allclose(ximg.xi.get_affines(),
                             {'scanner': new_affine2})
    ximg.xi.set_affines({'mni': new_affine})
    assert arr_dict_allclose(ximg.xi.get_affines(),
                             {'mni': new_affine, 'scanner': new_affine2})
    # Adjust affines gives the same reult back.
    ximg_adj = ximg.xi.with_updated_affines()
    assert arr_dict_allclose(ximg.xi.get_affines(),
                             {'mni': new_affine, 'scanner': new_affine2})
    # Coordinates with irregular spacing give affine errors.
    ximg_adj.coords['k'] = [0, 1, 3, 4, 5]
    with pytest.raises(XibFormatError,
                       match='Cannot handle irregular voxel spacing for "k"'):
        ximg_adj.xi.with_updated_affines()


def test_guess_format():
    root = Path('foo') / 'bar' / 'baz.suff'
    for v, exp in ((root, None),
                   (root.with_suffix('.nii'), None),
                   (root.with_suffix('.json'), 'bids'),
                   (root.with_suffix('.ximg'), 'zarr'),
                   (root.with_suffix('.nc'), 'netcdf'),
                   (root.with_suffix('.foo'), None)):

        assert PROCESSORS.guess_format(v) == exp
        assert PROCESSORS.guess_format(str(v)) == exp


def test__path2class():
    for url, exp_class in (
        ('/foo/bar/sub-07_T1w.nii.gz', nib.Nifti1Image),
        ('http://localhost:8999/sub-07_T1w.nii.gz', nib.Nifti1Image),
        ('/foo/bar/sub-07_T1w.nii', nib.Nifti1Image),
        ('http://localhost:8999/sub-07_T1w.nii', nib.Nifti1Image),
        ('/foo/bar/sub-07_T1w.mnc', nib.Minc1Image),
        ('http://localhost:8999/sub-07_T1w.mnc', nib.Minc1Image),
    ):
        assert _path2class(url) == exp_class


def test_drop_suffix():
    for inp, suffixes, exp_out in (
        ('foo/bar', ['.nii'], 'foo/bar'),
        ('foo/bar', '.nii', 'foo/bar'),
        ('foo/bar.baz', ['.nii'], 'foo/bar.baz'),
        ('foo/bar.nii', ['.nii'], 'foo/bar'),
        ('foo/bar.nii', '.nii', 'foo/bar'),
        ('foo/bar.nii.gz', ['.nii'], 'foo/bar.nii.gz'),
        ('foo/bar.nii.gz', ['.nii.gz', '.nii'], 'foo/bar'),
    ):
        assert drop_suffix(inp, suffixes) == exp_out
        assert drop_suffix(Path(inp), suffixes) == Path(exp_out)


def test_replace_suffix():
    for inp, suffixes, new_suffix, exp_out in (
        ('foo/bar', ['.nii'], '.json', 'foo/bar.json'),
        ('foo/bar', '.nii', '.json', 'foo/bar.json'),
        ('foo/bar.baz', ['.nii'], '.boo', 'foo/bar.boo'),
        ('foo/bar.nii', ['.nii'], '.boo', 'foo/bar.boo'),
        ('foo/bar.nii', '.nii', '.boo', 'foo/bar.boo'),
        ('foo/bar.nii.gz', ['.nii'], '.boo', 'foo/bar.nii.boo'),
        ('foo/bar.nii.gz', ['.nii.gz', '.nii'], '.boo', 'foo/bar.boo'),
    ):
        assert replace_suffix(inp, suffixes, new_suffix) == exp_out
        assert replace_suffix(Path(inp), suffixes, new_suffix) == Path(exp_out)


def test_json_attrs():
    # Test utilities to load / save JSON attrs
    d = {'foo': 1, 'bar': [2, 3]}
    assert _attrs2json_attrs(d) == d
    assert _json_attrs2attrs(d) == d
    dd = {'foo': 1, 'bar': {'baz': 4}}
    ddj = {'foo': 1, 'bar': ['__json__', '{"baz": 4}']}
    assert _attrs2json_attrs(dd) == ddj
    assert _json_attrs2attrs(ddj) == dd
    arr = rng.integers(0, 10, size=(3, 4)).tolist()
    arr_j = json.dumps(arr)
    dd = {'foo': 1, 'bar': {'baz': [2, 3]}, 'baf': arr}
    ddj = {'foo': 1,
           'bar': ['__json__', '{"baz": [2, 3]}'],
           'baf': ['__json__', arr_j]}
    assert _attrs2json_attrs(dd) == ddj
    assert _json_attrs2attrs(ddj) == dd
    ragged_arr = {'foo': {'bar': [[1], [2, 3]]}}
    raj = {'foo': ['__json__', '{"bar": [[1], [2, 3]]}']}
    assert _attrs2json_attrs(ragged_arr) == raj
    assert _json_attrs2attrs(raj) == ragged_arr


def test_jdumps():
    d = {'foo': 1, 'bar': [2, 3]}
    assert json.loads(_jdumps(d)) == d

    class C:
        pass

    bad_d = {'foo': C()}
    with pytest.raises(TypeError):
        _jdumps(bad_d)


def _check_dims_coords(ximg):
    ndims = len(ximg.dims)
    space_dims = ximg.dims[:3]
    assert space_dims == _NI_SPACE_DIMS
    assert space_dims == tuple(ximg.coords)[:3]
    for dim_no, dim in enumerate(space_dims):
        coord = ximg.coords[dim]
        assert np.all(np.array(coord) == np.arange(ximg.shape[dim_no]))
    if ndims > 3:
        assert ximg.dims[3] == _NI_TIME_DIM


@skip_without_file(JC_EG_FUNC)
def test_nib_loader_jc():
    img = nib.load(JC_EG_FUNC)
    ximg = load_nibabel(JC_EG_FUNC)
    assert ximg.attrs == JC_EG_FUNC_ATTRS
    _check_dims_coords(ximg)
    assert np.all(np.array(ximg) == img.get_fdata())


@skip_without_file(JH_EG_FUNC)
def test_nib_loader_jh():
    img = nib.load(JH_EG_FUNC)
    ximg = load_nibabel(JH_EG_FUNC)
    assert ximg.attrs == {'RepetitionTime': 2.5,
                          'xib-affines':
                          {'scanner': img.affine.tolist()}
                         }
    _check_dims_coords(ximg)


if fetcher.have_file(JC_EG_FUNC):
    img = nib.load(JC_EG_FUNC)
    JC_EG_FUNC_ATTRS = json.loads(JC_EG_FUNC_JSON.read_text())
    JC_EG_FUNC_ATTRS_RAW = {
        'xib-FrequencyEncodingDirection': 'i',
         'PhaseEncodingDirection': 'j',
         'SliceEncodingDirection': 'k',
         'RepetitionTime': 2.0,
         'xib-affines':
         {'scanner': img.affine.tolist()}
    }
    JC_EG_FUNC_ATTRS.update(JC_EG_FUNC_ATTRS_RAW)
else:  # For parametrized tests.
    JC_EG_FUNC_ATTRS = {}
    JC_EG_FUNC_ATTRS_RAW = {}


if fetcher.have_file(JC_EG_ANAT):
    img = nib.load(JC_EG_ANAT)
    JC_EG_ANAT_ATTRS = json.loads(JC_EG_ANAT_JSON.read_text())
    JC_EG_ANAT_ATTRS_RAW = {
        'xib-FrequencyEncodingDirection': 'j',
         'PhaseEncodingDirection': 'i',
         'SliceEncodingDirection': 'k',
         'xib-affines':
         {'scanner': img.affine.tolist()}
    }
    JC_EG_ANAT_ATTRS.update(JC_EG_ANAT_ATTRS_RAW)
else:  # For parametrized tests.
    JC_EG_ANAT_ATTRS = {}
    JC_EG_ANAT_ATTRS_RAW = {}



@skip_without_file(JC_EG_ANAT)
def test_anat_loader():
    img = nib.load(JC_EG_ANAT)
    for loader, in_path in product(
        (load, load_bids, load_nibabel),
        (JC_EG_ANAT, str(JC_EG_ANAT),
         JC_EG_ANAT_JSON, str(JC_EG_ANAT_JSON))):
        ximg = loader(in_path)
        assert ximg.shape == (176, 256, 256)
        assert ximg.dims == _NI_SPACE_DIMS
        assert ximg.name == JC_EG_ANAT.name.split('.')[0]
        assert ximg.attrs == JC_EG_ANAT_ATTRS
        _check_dims_coords(ximg)
        assert np.all(np.array(ximg) == img.get_fdata())


@skip_without_file(JC_EG_ANAT)
@skip_without_file(JC_EG_ANAT_JSON)
def test_anat_loader_http(fserver):
    nb_img = nib.load(JC_EG_ANAT)
    # Read nibabel from HTTP
    # Original gz
    name_gz = JC_EG_ANAT.name
    # Uncompressed, no gz
    name_no_gz = JC_EG_ANAT.with_suffix('').name
    out_path = fserver.server_path / name_no_gz
    nib.save(nb_img, out_path)
    for name in (name_gz, name_no_gz):
        out_url = fserver.make_url(name)
        ximg = load(out_url)
        # Check we can read the data
        ximg.compute()
        # Check parameters
        assert ximg.shape == (176, 256, 256)
        assert ximg.name == JC_EG_ANAT.name.split('.')[0]
        assert ximg.attrs == JC_EG_ANAT_ATTRS
        _check_dims_coords(ximg)
        assert np.all(np.array(ximg) == nb_img.get_fdata())
    # Refuse to load pair files.
    pair_path = fserver.server_path / 'pair.img'
    nib.save(nb_img, pair_path)
    json_path = fserver.server_path / 'pair.json'
    json_path.write_text(JC_EG_ANAT_JSON.read_text())
    for name in (pair_path.name, json_path.name):
        out_url = fserver.make_url(name)
        with pytest.raises(XibFileError):
            load(out_url)


@skip_without_file(JC_EG_ANAT)
def test_anat_loader_http_params(fserver, tmp_path):
    # Test params get passed through in kwargs.
    nb_img = nib.load(JC_EG_ANAT)
    out_url = 'simplecache::' + fserver.make_url(JC_EG_ANAT.name)
    out_cache = tmp_path / 'files'
    assert not out_cache.is_dir()
    ximg = load(out_url,
                simplecache={'cache_storage': str(out_cache)})
    assert np.all(np.array(ximg) == nb_img.get_fdata())
    assert out_cache.is_dir()
    assert len(list(out_cache.glob('*'))) == 2  # JSON and Nifti


@skip_without_file(JC_EG_ANAT)
def test_round_trip(tmp_path):
    ximg = load(JC_EG_ANAT)
    assert ximg.shape == (176, 256, 256)
    _check_dims_coords(ximg)
    out_path = tmp_path / 'out.ximg'
    save(ximg, out_path)
    back = load(out_path)
    assert back.shape == (176, 256, 256)
    assert back.attrs == JC_EG_ANAT_ATTRS
    _check_dims_coords(back)
    # And again
    save(ximg, out_path)
    back = load(out_path)
    assert back.attrs == JC_EG_ANAT_ATTRS
    _check_dims_coords(back)
    # With url
    back = load(f'file:///{out_path}')
    assert back.attrs == JC_EG_ANAT_ATTRS
    _check_dims_coords(back)


@skip_without_file(JC_EG_ANAT)
def test_rt_header(tmp_path):
    # Modifying header modifies ximg and out
    in_path = tmp_path / 'in.nii'
    out_path = tmp_path / 'out.nii'
    nimg = nib.load(JC_EG_ANAT)
    old_affine = nimg.affine.copy()
    new_affine = np.diag([2.1, 3.2, 4.1, 1])
    new_affine[:3, 3] = [11, 12, 13]
    nib.save(nimg, in_path)
    ximg = load(in_path)
    assert arr_dict_allclose(ximg.xi.get_affines(), {'scanner': old_affine})
    save(ximg, out_path)
    ximg = load(out_path)
    assert arr_dict_allclose(ximg.xi.get_affines(), {'scanner': old_affine})
    nimg.set_sform(new_affine, 'scanner')
    nib.save(nimg, in_path)
    ximg = load(in_path)
    assert arr_dict_allclose(ximg.xi.get_affines(), {'scanner': new_affine})
    save(ximg, out_path)
    ximg = load(out_path)
    assert arr_dict_allclose(ximg.xi.get_affines(), {'scanner': new_affine})
    nimg.set_sform(new_affine, 'aligned')
    nib.save(nimg, in_path)
    ximg = load(in_path)
    assert arr_dict_allclose(ximg.xi.get_affines(),
                             {'aligned': new_affine, 'scanner': old_affine})
    save(ximg, out_path)
    ximg = load(out_path)
    assert arr_dict_allclose(ximg.xi.get_affines(),
                             {'aligned': new_affine, 'scanner': old_affine})
    nimg.set_sform(new_affine, 'template')
    nib.save(nimg, in_path)
    ximg = load(in_path)
    assert arr_dict_allclose(ximg.xi.get_affines(),
                             {'template': new_affine, 'scanner': old_affine})
    save(ximg, out_path)
    ximg = load(out_path)
    assert arr_dict_allclose(ximg.xi.get_affines(),
                             {'template': new_affine, 'scanner': old_affine})
    nimg.set_qform(old_affine, 'mni')
    nimg.set_sform(new_affine, 'aligned')
    nib.save(nimg, in_path)
    ximg = load(in_path)
    assert arr_dict_allclose(ximg.xi.get_affines(),
                             {'mni': old_affine, 'aligned': new_affine})
    save(ximg, out_path)
    ximg = load(out_path)
    assert arr_dict_allclose(ximg.xi.get_affines(),
                             {'mni': old_affine, 'aligned': new_affine})


@h5netcdf_test
@skip_without_file(JC_EG_ANAT)
def test_round_trip_netcdf(tmp_path):
    ximg = load(JC_EG_ANAT)
    out_path = tmp_path / 'out.nc'
    save(ximg, out_path)
    back = load(out_path)
    assert back.shape == (176, 256, 256)
    assert back.attrs == JC_EG_ANAT_ATTRS
    back = load(f'file:///{out_path}')
    assert back.attrs == JC_EG_ANAT_ATTRS
    _check_dims_coords(back)


@skip_without_file(JC_EG_ANAT)
@skip_without_file(JC_EG_FUNC)
@pytest.mark.parametrize("img_path", [JC_EG_ANAT, JC_EG_FUNC])
def test_affines(img_path):
    ximg = load(img_path)
    nib_img = nib.load(img_path)
    affines = ximg.xi.get_affines()
    assert list(affines) == ['scanner']
    assert np.all(affines['scanner'] == nib_img.affine)
    back = ximg.xi.with_updated_affines()
    assert np.all(back.xi.get_affines()['scanner'] == nib_img.affine)
    xT = ximg.T
    sp_dims = _NI_SPACE_DIMS
    assert xT.dims == ximg.dims[::-1]
    transposed = xT.xi.with_updated_affines()
    assert np.all(transposed.xi.get_affines()['scanner'] == nib_img.affine)
    for i, (name, slice_no) in enumerate(zip(sp_dims, (42, 32, 10))):
        ximg_plane = ximg.sel(**{name: slice_no})
        assert ximg_plane.dims == tuple(d for d in ximg.dims if d != name)
        assert tuple(ximg_plane.coords) == ximg.dims
        adj_img = ximg_plane.xi.with_updated_affines()
        new_origin = np.array([0, 0, 0, 1])
        new_origin[i] = slice_no
        new_affine = nib_img.affine.copy()
        new_affine[:, 3] = nib_img.affine @ new_origin
        assert np.all(adj_img.xi.get_affines()['scanner'] == new_affine)
        # Check Nibabel slicer gives the same result.
        slicers = [slice(None) for d in range(len(ximg.dims))]
        slicers[i] = slice(slice_no, slice_no + 1)
        assert np.all(nib_img.slicer[tuple(slicers)].affine == new_affine)
    for i, (name, spacing) in enumerate(zip(sp_dims, (2, 3, 4))):
        slicers = [slice(None) for d in range(len(ximg.dims))]
        slicers[i] = slice(None, None, spacing)
        ximg_sliced = ximg[tuple(slicers)]
        assert ximg_sliced.dims == ximg.dims
        assert tuple(ximg_sliced.coords) == ximg.dims
        adj_img = ximg_sliced.xi.with_updated_affines()
        new_affine = nib_img.affine.copy()
        scalers = np.ones(3)
        scalers[i] = spacing
        new_affine[:3, :3] = nib_img.affine[:3, :3] @ np.diag(scalers)
        assert np.all(adj_img.xi.get_affines()['scanner'] == new_affine)
        # Check Nibabel slicer gives the same result.
        assert np.all(nib_img.slicer[tuple(slicers)].affine == new_affine)
    # Some more complex slicings, benchmark against nibabel
    for slicers in permutations([slice(10, 20, 2),
                                 slice(30, 15, -1),
                                 slice(5, 16, 3)]):
        sliced_ximg = ximg[tuple(slicers)].xi.with_updated_affines()
        sliced_nib_img = nib_img.slicer[tuple(slicers)]
        assert np.all(sliced_ximg.xi.get_affines()['scanner'] ==
                    sliced_nib_img.affine)


def test_tornado(fserver):
    # Test static file server for URL reads
    fserver.write_text_to('text_file', 'some text')
    fserver.write_bytes_to('binary_file', b'binary')
    response = fserver.get('text_file')
    assert response.status_code == 200
    assert response.text == 'some text'
    assert fserver.read_text('text_file') == 'some text'
    assert fserver.read_bytes('text_file') == b'some text'
    assert fserver.read_bytes('binary_file') == b'binary'


@h5netcdf_test
@skip_without_file(JC_EG_ANAT)
def test_round_trip_netcdf_url(fserver):
    ximg = load(JC_EG_ANAT)
    save(ximg, fserver.server_path / 'out.nc')
    out_url = fserver.make_url('out.nc')
    back = load(out_url)
    assert back.shape == (176, 256, 256)
    assert back.attrs == JC_EG_ANAT_ATTRS
    _check_dims_coords(back)


@skip_without_file(JC_EG_ANAT)
def test_matching_img_error(tmp_path):
    out_json = tmp_path / JC_EG_ANAT_JSON.name
    with pytest.raises(XibFileError, match='does not appear to exist'):
        load(out_json)
    shutil.copy2(JC_EG_ANAT_JSON, tmp_path)
    with pytest.raises(XibFileError, match='No valid file matching'):
        load(out_json)
    out_img = tmp_path / JC_EG_ANAT.name
    shutil.copy2(JC_EG_ANAT, tmp_path)
    back = load(out_img)
    assert back.shape == (176, 256, 256)
    assert back.attrs == JC_EG_ANAT_ATTRS
    _check_dims_coords(back)
    os.unlink(out_img)
    with pytest.raises(XibFileError, match='does not appear to exist'):
        load(out_img)
    shutil.copy2(JC_EG_ANAT, tmp_path)
    os.unlink(out_json)
    back = load(out_img)
    _check_dims_coords(back)
    assert back.attrs == JC_EG_ANAT_ATTRS_RAW
    back = load_bids(out_img, require_json=False)
    assert back.attrs == JC_EG_ANAT_ATTRS_RAW
    _check_dims_coords(back)
    with pytest.raises(XibFileError, match='`require_json` is True'):
        load_bids(out_img, require_json=True)


def test_ni_sort_expand_dims():
    assert _ni_sort_expand_dims([]) == ([],
                                        ['i', 'j', 'k'],
                                        [0, 1, 2])
    assert _ni_sort_expand_dims(['time']) == (['time'],
                                              ['i', 'j', 'k'],
                                              [0, 1, 2])
    assert (_ni_sort_expand_dims(['j']) ==
            (['j'],
             ['i', 'k'],
             [0, 2]))
    assert (_ni_sort_expand_dims(['time', 'j']) ==
            (['j', 'time'],
             ['i', 'k'],
             [0, 2]))
    assert (_ni_sort_expand_dims(['j', 'i']) ==
            (['i', 'j'],
             ['k'],
             [2]))
    assert (_ni_sort_expand_dims(['time', 'j', 'k']) ==
            (['j', 'k', 'time'],
             ['i'],
             [0]))


@skip_without_file(JC_EG_ANAT)
@skip_without_file(JC_EG_FUNC)
@pytest.mark.parametrize("img_path, attrs_raw",
                         ((JC_EG_ANAT, JC_EG_ANAT_ATTRS_RAW),
                          (JC_EG_FUNC, JC_EG_FUNC_ATTRS_RAW)))
def test_to_nifti(img_path, attrs_raw):
    orig_img = nib.load(img_path)
    orig_data = orig_img.get_fdata()
    ximg = load(img_path)
    # Check data is the same for basic load.
    assert np.all(ximg == orig_data)
    # Basic conversion.
    img, attrs = to_nifti(ximg)
    assert np.all(img.get_fdata() == orig_data)
    assert arr_dict_allclose(hdr2attrs(img.header), attrs_raw)
    img, attrs = to_nifti(ximg.T)
    assert np.all(img.get_fdata() == orig_data)
    assert arr_dict_allclose(hdr2attrs(img.header), attrs_raw)
    img, attrs = to_nifti(ximg.T.sel(k=32))  # Drop k axis
    assert np.all(img.get_fdata() == orig_data[:, :, 32:33])
    # This changes the origin of the affine.
    new_affine = orig_img.slicer[:, :, 32:33].affine
    exp_attrs_32 = deepcopy(attrs_raw)
    exp_attrs_32['xib-affines']['scanner'] = new_affine
    assert arr_dict_allclose(hdr2attrs(img.header), exp_attrs_32)


@skip_without_file(JC_EG_ANAT)
@skip_without_file(JC_EG_FUNC)
@pytest.mark.parametrize("img_path, attrs",
                         ((JC_EG_ANAT, JC_EG_ANAT_ATTRS),
                          (JC_EG_FUNC, JC_EG_FUNC_ATTRS)))
def test_to_bids(img_path, attrs, tmp_path):
    nib_img = nib.load(img_path)
    fdata = nib_img.get_fdata()
    ximg = load(img_path)
    out_fname = tmp_path / 'out.json'
    save(ximg, out_fname)
    back_nib = nib.load(tmp_path / 'out.nii.gz')
    assert np.allclose(back_nib.get_fdata(), fdata)
    with open(out_fname, 'rt') as fobj:
        back_attrs = json.load(fobj)
    assert arr_dict_allclose(back_attrs, attrs)
    assert arr_dict_allclose(load(out_fname).attrs, ximg.attrs)
    out_fname = tmp_path / 'out2.nii'
    save(ximg, out_fname)
    back_nib = nib.load(out_fname)
    assert np.allclose(back_nib.get_fdata(), fdata)
    with open(tmp_path / 'out2.json', 'rt') as fobj:
        back_attrs = json.load(fobj)
    assert arr_dict_allclose(back_attrs, attrs)
    assert arr_dict_allclose(load(out_fname).attrs, ximg.attrs)


def test_loadsave_errors(tmp_path):
    ximg = from_array(np.arange(24).reshape((2, 3, 4)))
    save(ximg, tmp_path / 'ok.nii')
    # Nibabel, wrong format.
    with pytest.raises(nib.filebasedimages.ImageFileError):
        save(ximg, tmp_path / 'bad1.foo')
    # Nibabel, forced wrong format.
    with pytest.raises(XibFormatError):
        save(ximg, tmp_path / 'bad1.nii', format='foo')
    # Nibabel, forced wrong format for load.
    with pytest.raises(XibFormatError):
        save(ximg, tmp_path / 'ok.nii', format='foo')



""" Test scripting of  GLM
"""

import numpy as np

import pandas as pd

import xarray as xr

import nibabel as nib
import xibabel as xib

from xibabel import testing
from xibabel.tests.markers import nipy_test

import pytest

img_path_root = (testing.DATA_PATH /
                'ds000105' /
                'sub-1' /
                'func' /
                'sub-1_task-objectviewing_run-01_')

bold_path = img_path_root.with_name(img_path_root.name + 'bold.nii.gz')
tsv_path = img_path_root.with_name(img_path_root.name + 'events.tsv')


@nipy_test
@pytest.mark.skipif(not bold_path.is_file(), reason=f'Missing "{bold_path}"')
@pytest.mark.skipif(not tsv_path.is_file(), reason=f'Missing "{tsv_path}"')
def test_glm():

    # For constructing the design.
    from nipy.modalities.fmri.design import block_design, natural_spline

    # Load the events ready to make a design.
    event_df = pd.read_csv(tsv_path, sep='\t')
    df = event_df.rename(columns={'onset': 'start', 'duration': 'end'})
    df['end'] = df['start'] + df['end']
    block_spec = df.to_records(index=None)

    nib_img = nib.load(bold_path)
    vol_shape = nib_img.shape[:-1]
    n_vols = nib_img.shape[-1]
    TR = nib_img.header.get_zooms()[-1]
    # Of course this array comes naturally from xirr['time'] below.
    t = np.arange(n_vols) * TR
    regressors, contrasts = block_design(block_spec, t)
    con_tt0 = contrasts['trial_type_0']
    n_contrasts = con_tt0.shape[0]
    # Add drift regressors.
    drift = natural_spline(t)
    n_drift = drift.shape[1]
    design_matrix = np.hstack([regressors, drift])
    # Contrasts for the eight event types.
    con_mat = np.hstack([con_tt0, np.zeros((n_contrasts, n_drift))])

    # For notation!
    X = design_matrix

    # Analysis the old way.
    # The 4D array (i, j, k, time)
    data = nib_img.get_fdata()

    # Reshape to time by (i * j * k)
    vols_by_voxels = np.reshape(data, (-1, n_vols)).T

    # Run estimation with B = X+ Y
    # To get B such that errors (E = Y - X @ B) are minimized in the sense
    # of having the smallest np.sum(E ** 2, axis=0) for each column.
    # Betas for each voxel (n_cols_in_X by n_voxels).
    pX = np.linalg.pinv(X)
    B = pX @ vols_by_voxels

    # Contrast applied.  Two slopes compared
    c = con_mat[4, :]
    con_vec = c @ B

    con_arr = np.reshape(con_vec, vol_shape)

    # OK - that was the crappy way.
    # Now the awesome way.
    xib_img = xib.load(bold_path)
    # assert isinstance(xib_img, xr.DataArray)

    # Make the design
    xesign = xr.DataArray(pX, dims=['p', 'time'])

    # Optional: make the data chunky.
    chunked = xib_img.chunk({'k': 5})

    # Do the estimation
    xB = xr.dot(xesign, chunked, dim=['time', 'time'])

    xC = xr.DataArray(c, dims=['p'])

    x_c_arr = xr.dot(xC, xB, dim=['p', 'p'])

    assert np.allclose(x_c_arr, con_arr)



""" Testing testing (... testing).
"""

import numpy as np

import xibabel.testing as xit


def test_assert_dict_allclose():
    assert xit.arr_dict_allclose({}, {})
    assert not xit.arr_dict_allclose({}, 1)
    assert xit.arr_dict_allclose({'foo': 1}, {'foo': 1})
    assert not xit.arr_dict_allclose({'foo': 1}, {'bar': 1})
    assert not xit.arr_dict_allclose({'foo': 1}, {'foo': 2})
    assert not xit.arr_dict_allclose({'foo': 1}, {'foo': 1, 'bar': 2})
    assert xit.arr_dict_allclose({'foo': np.array([1, 2, 3])},
                                {'foo': np.array([1, 2, 3])})
    assert not xit.arr_dict_allclose({'foo': np.array([1, 2, 3])},
                                    {'bar': np.array([2, 2, 3])})
    # Test atol
    assert xit.arr_dict_allclose({'foo': np.array([1, 2, 0])},
                                {'foo': np.array([1, 2, 1.1e-9])})
    assert not xit.arr_dict_allclose({'foo': np.array([1, 2, 0])},
                                    {'foo': np.array([1, 2, 1.1e-8])})
    assert xit.arr_dict_allclose({'foo': np.array([1, 2, 0])},
                                {'foo': np.array([1, 2, 1.1e-8])},
                                atol=1e-7)
    # Also works with lists.
    assert xit.arr_dict_allclose({'foo': [1, 2, 0]},
                                 {'foo': [1, 2, 1.1e-9]})
    # Test rtol
    assert xit.arr_dict_allclose({'foo': np.array([1, 2, 10])},
                                {'foo': np.array([1, 2, 10 + 1.1e-5])})
    assert not xit.arr_dict_allclose({'foo': np.array([1, 2, 10])},
                                    {'foo': np.array([1, 2, 10 + 1.1e-4])})
    assert xit.arr_dict_allclose({'foo': np.array([1, 2, 10])},
                                {'foo': np.array([1, 2, 10 + 1.1e-4])},
                                rtol=1e-3)
    # Recursive.
    assert xit.arr_dict_allclose({'foo': {'bar': np.array([1, 2, 3])}},
                                {'foo': {'bar': np.array([1, 2, 3])}})
    assert not xit.arr_dict_allclose({'foo': {'bar': np.array([1, 2, 3])}},
                                    {'foo': {'bar': np.array([1, 2, 3.1])}})

    # Different types
    assert not xit.arr_dict_allclose({'foo': ['bar', 'boo', 'baz']},
                                     {'foo': np.array([1, 2, 3.1])})
